@@ -13,6 +13,8 @@
 (*                                                                        *)
 (**************************************************************************)
 
+let shell = "/bin/sh"
+
 type error =
     E2BIG
   | EACCES
@@ -203,72 +205,13 @@ type wait_flag =
 external execv : string -> string array -> 'a = "unix_execv"
 external execve : string -> string array -> string array -> 'a = "unix_execve"
 external execvp : string -> string array -> 'a = "unix_execvp"
-external execvpe_c :
-            string -> string array -> string array -> 'a = "unix_execvpe"
-
-let execvpe_ml name args env =
-  (* Try to execute the given file *)
-  let exec file =
-    try
-      execve file args env
-    with Unix_error(ENOEXEC, _, _) ->
-      (* Assume this is a script and try to execute through the shell *)
-      let argc = Array.length args in
-      (* Drop the original args.(0) if it is there *)
-      let new_args = Array.append
-        [| "/bin/sh"; file |]
-        (if argc = 0 then args else Array.sub args 1 (argc - 1)) in
-      execve new_args.(0) new_args env in
-  (* Try each path element in turn *)
-  let rec scan_dir eacces = function
-  | [] ->
-      (* No matching file was found (if [eacces = false]) or
-         a matching file was found but we got a "permission denied"
-         error while trying to execute it (if [eacces = true]).
-         Raise the error appropriate to each case. *)
-      raise (Unix_error((if eacces then EACCES else ENOENT),
-                        "execvpe", name))
-  | dir :: rem ->
-      let dir =  (* an empty path element means the current directory *)
-        if dir = "" then Filename.current_dir_name else dir in
-      try
-        exec (Filename.concat dir name)
-      with Unix_error(err, _, _) as exn ->
-        match err with
-        (* The following errors are treated as nonfatal, meaning that
-           we will ignore them and continue searching in the path.
-           Among those errors, EACCES is recorded specially so as
-           to produce the correct exception in the end.
-           To determine which errors are nonfatal, we looked at the
-           execvpe() sources in Glibc and in OpenBSD. *)
-        | EACCES ->
-            scan_dir true rem
-        | EISDIR|ELOOP|ENAMETOOLONG|ENODEV|ENOENT|ENOTDIR|ETIMEDOUT ->
-            scan_dir eacces rem
-        (* Other errors, e.g. E2BIG, are fatal and abort the search. *)
-        | _ ->
-            raise exn in
-  if String.contains name '/' then
-    (* If the command name contains "/" characters, don't search in path *)
-    exec name
-  else
-    (* Split path into elements and search in these elements *)
-    (try unsafe_getenv "PATH" with Not_found -> "/bin:/usr/bin")
-    |> String.split_on_char ':'
-    |> scan_dir false
-      (* [unsafe_getenv] and not [getenv] to be consistent with [execvp],
-         which looks up the PATH environment variable whether SUID or not. *)
-
-let execvpe name args env =
-  try  
-    execvpe_c name args env
-  with Unix_error(ENOSYS, _, _) ->
-    execvpe_ml name args env
+external execvpe : string -> string array -> string array -> 'a = "unix_execvpe"
 
 external fork : unit -> int = "unix_fork"
 external wait : unit -> int * process_status = "unix_wait"
 external waitpid : wait_flag list -> int -> int * process_status
    = "unix_waitpid"
+external _exit : int -> 'a = "unix_exit"
 external getpid : unit -> int = "unix_getpid"
 external getppid : unit -> int = "unix_getppid"
 external nice : int -> int = "unix_nice"
@@ -301,8 +244,8 @@ type file_perm = int
 
 external openfile : string -> open_flag list -> file_perm -> file_descr
            = "unix_open"
-
 external close : file_descr -> unit = "unix_close"
+external fsync : file_descr -> unit = "unix_fsync"
 external unsafe_read : file_descr -> bytes -> int -> int -> int
    = "unix_read"
 external unsafe_write : file_descr -> bytes -> int -> int -> int = "unix_write"
@@ -332,9 +275,9 @@ let single_write_substring fd buf ofs len =
   single_write fd (Bytes.unsafe_of_string buf) ofs len
 
 external in_channel_of_descr : file_descr -> in_channel
-                             = "caml_ml_open_descriptor_in"
+                             = "unix_inchannel_of_filedescr"
 external out_channel_of_descr : file_descr -> out_channel
-                              = "caml_ml_open_descriptor_out"
+                              = "unix_outchannel_of_filedescr"
 external descr_of_in_channel : in_channel -> file_descr
                              = "caml_channel_descriptor"
 external descr_of_out_channel : out_channel -> file_descr
@@ -378,7 +321,7 @@ external fstat : file_descr -> stats = "unix_fstat"
 external isatty : file_descr -> bool = "unix_isatty"
 external unlink : string -> unit = "unix_unlink"
 external rename : string -> string -> unit = "unix_rename"
-external link : string -> string -> unit = "unix_link"
+external link : ?follow:bool -> string -> string -> unit = "unix_link"
 
 module LargeFile =
   struct
@@ -495,8 +438,10 @@ type tm =
     tm_yday : int;
     tm_isdst : bool }
 
-external time : unit -> float = "unix_time"
-external gettimeofday : unit -> float = "unix_gettimeofday"
+external time : unit -> (float [@unboxed]) =
+  "unix_time" "unix_time_unboxed" [@@noalloc]
+external gettimeofday : unit -> (float [@unboxed]) =
+  "unix_gettimeofday" "unix_gettimeofday_unboxed" [@@noalloc]
 external gmtime : float -> tm = "unix_gmtime"
 external localtime : float -> tm = "unix_localtime"
 external mktime : tm -> float * tm = "unix_mktime"
@@ -597,7 +542,7 @@ type msg_flag =
   | MSG_DONTROUTE
   | MSG_PEEK
 
-external socket : 
+external socket :
   ?cloexec: bool -> socket_domain -> socket_type -> int -> file_descr
   = "unix_socket"
 external socketpair :
@@ -659,6 +604,7 @@ type socket_bool_option =
   | SO_ACCEPTCONN
   | TCP_NODELAY
   | IPV6_ONLY
+  | SO_REUSEPORT
 
 type socket_int_option =
     SO_SNDBUF
@@ -861,7 +807,7 @@ let getnameinfo_emulation addr opts =
           let kind = if List.mem NI_DGRAM opts then "udp" else "tcp" in
           (getservbyport p kind).s_name
         with Not_found ->
-          string_of_int p in
+          Int.to_string p in
       { ni_hostname = hostname; ni_service = service }
 
 let getnameinfo addr opts =
@@ -936,67 +882,48 @@ let rec waitpid_non_intr pid =
   try waitpid [] pid
   with Unix_error (EINTR, _, _) -> waitpid_non_intr pid
 
-external sys_exit : int -> 'a = "caml_sys_exit"
+external spawn : string -> string array -> string array option ->
+                 bool -> int array -> int
+               = "unix_spawn"
 
 let system cmd =
-  match fork() with
-     0 -> begin try
-            execv "/bin/sh" [| "/bin/sh"; "-c"; cmd |]
-          with _ ->
-            sys_exit 127
-          end
-  | id -> snd(waitpid_non_intr id)
+  let pid = spawn shell [| shell; "-c"; cmd |] None false [| 0; 1; 2 |] in
+  snd(waitpid_non_intr pid)
 
-(* Duplicate [fd] if needed to make sure it isn't one of the
-   standard descriptors (stdin, stdout, stderr).
-   Note that this function always leaves the standard descriptors open,
-   the caller must take care of closing them if needed.
-   The "cloexec" mode doesn't matter, because
-   the descriptor returned by [dup] will be closed before the [exec],
-   and because no other thread is running concurrently
-   (we are in the child process of a fork).
- *)
-let rec file_descr_not_standard fd =
-  if fd >= 3 then fd else file_descr_not_standard (dup fd)
-
-let safe_close fd =
-  try close fd with Unix_error(_,_,_) -> ()
-
-let perform_redirections new_stdin new_stdout new_stderr =
-  let new_stdin = file_descr_not_standard new_stdin in
-  let new_stdout = file_descr_not_standard new_stdout in
-  let new_stderr = file_descr_not_standard new_stderr in
-  (*  The three dup2 close the original stdin, stdout, stderr,
-      which are the descriptors possibly left open
-      by file_descr_not_standard *)
-  dup2 ~cloexec:false new_stdin stdin;
-  dup2 ~cloexec:false new_stdout stdout;
-  dup2 ~cloexec:false new_stderr stderr;
-  safe_close new_stdin;
-  safe_close new_stdout;
-  safe_close new_stderr
+let create_process_gen usepath cmd args optenv
+                       new_stdin new_stdout new_stderr =
+  let toclose = ref [] in
+  let close_after () =
+    List.iter
+      (fun fd -> try close fd with Unix_error(_,_,_) -> ())
+      !toclose in
+  (* Duplicate [fd] if needed to make sure it isn't one of the
+     standard descriptors (stdin, stdout, stderr).
+     The temporary file descriptors created here will be closed
+     after the spawn, both in the parent (call to [close_after] below)
+     and in the child (they are close-on-exec). *)
+  let rec file_descr_not_standard fd =
+    if fd >= 3 then fd else begin
+      let fd' = dup ~cloexec:true fd in
+      toclose := fd' :: !toclose;
+      file_descr_not_standard fd'
+    end in
+  (* As an optimization, if a standard descriptor is not redirected,
+     i.e. "redirected to itself", don't duplicate it: the [unix_spawn]
+     C stub will perform no redirection either. *)
+  let redirections = [|
+    (if new_stdin = 0 then 0 else file_descr_not_standard new_stdin);
+    (if new_stdout = 1 then 1 else file_descr_not_standard new_stdout);
+    (if new_stderr = 2 then 2 else file_descr_not_standard new_stderr)
+  |] in
+  Fun.protect ~finally:close_after
+    (fun () -> spawn cmd args optenv usepath redirections)
 
 let create_process cmd args new_stdin new_stdout new_stderr =
-  match fork() with
-    0 ->
-      begin try
-        perform_redirections new_stdin new_stdout new_stderr;
-        execvp cmd args
-      with _ ->
-        sys_exit 127
-      end
-  | id -> id
+  create_process_gen true cmd args None new_stdin new_stdout new_stderr
 
 let create_process_env cmd args env new_stdin new_stdout new_stderr =
-  match fork() with
-    0 ->
-      begin try
-        perform_redirections new_stdin new_stdout new_stderr;
-        execvpe cmd args env
-      with _ ->
-        sys_exit 127
-      end
-  | id -> id
+  create_process_gen true cmd args (Some env) new_stdin new_stdout new_stderr
 
 type popen_process =
     Process of in_channel * out_channel
@@ -1006,26 +933,17 @@ type popen_process =
 
 let popen_processes = (Hashtbl.create 7 : (popen_process, int) Hashtbl.t)
 
-let open_proc cmd envopt proc input output error =
-   match fork() with
-     0 -> perform_redirections input output error;
-          let shell = "/bin/sh" in
-          let argv = [| shell; "-c"; cmd |] in
-          begin try
-            match envopt with
-            | Some env -> execve shell argv env
-            | None     -> execv shell argv
-          with _ ->
-            sys_exit 127
-          end
-   | id -> Hashtbl.add popen_processes proc id
+let open_proc prog args envopt proc input output error =
+  let pid =
+    create_process_gen false prog args envopt input output error in
+  Hashtbl.add popen_processes proc pid
 
-let open_process_in cmd =
+let open_process_args_in prog args =
   let (in_read, in_write) = pipe ~cloexec:true () in
   let inchan = in_channel_of_descr in_read in
   begin
     try
-      open_proc cmd None (Process_in inchan) stdin in_write stderr
+      open_proc prog args None (Process_in inchan) stdin in_write stderr
     with e ->
       close_in inchan;
       close in_write;
@@ -1034,12 +952,12 @@ let open_process_in cmd =
   close in_write;
   inchan
 
-let open_process_out cmd =
+let open_process_args_out prog args =
   let (out_read, out_write) = pipe ~cloexec:true () in
   let outchan = out_channel_of_descr out_write in
   begin
     try
-      open_proc cmd None (Process_out outchan) out_read stdout stderr
+      open_proc prog args None (Process_out outchan) out_read stdout stderr
     with e ->
     close_out outchan;
     close out_read;
@@ -1048,7 +966,7 @@ let open_process_out cmd =
   close out_read;
   outchan
 
-let open_process cmd =
+let open_process_args prog args =
   let (in_read, in_write) = pipe ~cloexec:true () in
   let (out_read, out_write) =
     try pipe ~cloexec:true ()
@@ -1057,7 +975,8 @@ let open_process cmd =
   let outchan = out_channel_of_descr out_write in
   begin
     try
-      open_proc cmd None (Process(inchan, outchan)) out_read in_write stderr
+      open_proc prog args None
+                (Process(inchan, outchan)) out_read in_write stderr
     with e ->
       close out_read; close out_write;
       close in_read; close in_write;
@@ -1067,7 +986,7 @@ let open_process cmd =
   close in_write;
   (inchan, outchan)
 
-let open_process_full cmd env =
+let open_process_args_full prog args env =
   let (in_read, in_write) = pipe ~cloexec:true () in
   let (out_read, out_write) =
     try pipe ~cloexec:true ()
@@ -1081,12 +1000,12 @@ let open_process_full cmd env =
   let errchan = in_channel_of_descr err_read in
   begin
     try
-      open_proc cmd (Some env) (Process_full(inchan, outchan, errchan))
+      open_proc prog args (Some env) (Process_full(inchan, outchan, errchan))
                 out_read in_write err_write
     with e ->
       close out_read; close out_write;
       close in_read; close in_write;
-      close err_read; close err_write; 
+      close err_read; close err_write;
       raise e
   end;
   close out_read;
@@ -1094,36 +1013,64 @@ let open_process_full cmd env =
   close err_write;
   (inchan, outchan, errchan)
 
+let open_process_shell fn cmd =
+  fn shell [|shell; "-c"; cmd|]
+let open_process_in cmd =
+  open_process_shell open_process_args_in cmd
+let open_process_out cmd =
+  open_process_shell open_process_args_out cmd
+let open_process cmd =
+  open_process_shell open_process_args cmd
+let open_process_full cmd =
+  open_process_shell open_process_args_full cmd
+
 let find_proc_id fun_name proc =
   try
-    let pid = Hashtbl.find popen_processes proc in
-    Hashtbl.remove popen_processes proc;
-    pid
+    Hashtbl.find popen_processes proc
   with Not_found ->
     raise(Unix_error(EBADF, fun_name, ""))
 
+let remove_proc_id proc =
+  Hashtbl.remove popen_processes proc
+
+let process_in_pid inchan =
+  find_proc_id "process_in_pid" (Process_in inchan)
+let process_out_pid outchan =
+  find_proc_id "process_out_pid" (Process_out outchan)
+let process_pid (inchan, outchan) =
+  find_proc_id "process_pid" (Process(inchan, outchan))
+let process_full_pid (inchan, outchan, errchan) =
+  find_proc_id "process_full_pid"
+    (Process_full(inchan, outchan, errchan))
+
 let close_process_in inchan =
-  let pid = find_proc_id "close_process_in" (Process_in inchan) in
+  let proc = Process_in inchan in
+  let pid = find_proc_id "close_process_in" proc in
+  remove_proc_id proc;
   close_in inchan;
   snd(waitpid_non_intr pid)
 
 let close_process_out outchan =
-  let pid = find_proc_id "close_process_out" (Process_out outchan) in
+  let proc = Process_out outchan in
+  let pid = find_proc_id "close_process_out" proc in
+  remove_proc_id proc;
   (* The application may have closed [outchan] already to signal
      end-of-input to the process.  *)
   begin try close_out outchan with Sys_error _ -> () end;
   snd(waitpid_non_intr pid)
 
 let close_process (inchan, outchan) =
-  let pid = find_proc_id "close_process" (Process(inchan, outchan)) in
+  let proc = Process(inchan, outchan) in
+  let pid = find_proc_id "close_process" proc in
+  remove_proc_id proc;
   close_in inchan;
   begin try close_out outchan with Sys_error _ -> () end;
   snd(waitpid_non_intr pid)
 
 let close_process_full (inchan, outchan, errchan) =
-  let pid =
-    find_proc_id "close_process_full"
-                 (Process_full(inchan, outchan, errchan)) in
+  let proc = Process_full(inchan, outchan, errchan) in
+  let pid = find_proc_id "close_process_full" proc in
+  remove_proc_id proc;
   close_in inchan;
   begin try close_out outchan with Sys_error _ -> () end;
   close_in errchan;
@@ -1158,8 +1105,8 @@ let establish_server server_fun sockaddr =
     (* The "double fork" trick, the process which calls server_fun will not
        leave a zombie process *)
     match fork() with
-       0 -> if fork() <> 0 then sys_exit 0;
-                                (* The son exits, the grandson works *)
+       0 -> if fork() <> 0 then _exit 0;
+                                (* The child exits, the grandchild works *)
             close sock;
             let inchan = in_channel_of_descr s in
             let outchan = out_channel_of_descr s in
@@ -1168,5 +1115,5 @@ let establish_server server_fun sockaddr =
                have done it already, and we are about to exit anyway
                (PR#3794) *)
             exit 0
-    | id -> close s; ignore(waitpid_non_intr id) (* Reclaim the son *)
+    | id -> close s; ignore(waitpid_non_intr id) (* Reclaim the child *)
   done
